﻿using System;
using System.Collections.Generic;
using System.Linq;
using Niche.GroupInspector.Domain;
using Niche.GroupInspector.Presentation;
using NUnit.Framework;
using Rhino.Mocks;

namespace Niche.GroupInspector.Tests
{
    [TestFixture]
    public class GroupInspectorPresenterTests
    {
        [SetUp]
        public void SetUp()
        {
            mRepository = new MockRepository();
        }

        [Test]
        public void Constructor_missingModel_throwsException()
        {
            Assert.Throws<ArgumentNullException>(
                () => new GroupInspectorPresenter(null, ViewStub()));
        }

        [Test]
        public void Constructor_givenModel_setsProperty()
        {
            var model = ModelStub();
            var view = ViewStub();
            var presenter = new GroupInspectorPresenter(model, view);
            Assert.That(presenter.Model, Is.EqualTo(model));
        }

        [Test]
        public void Constructor_missingView_throwsException()
        {
            Assert.Throws<ArgumentNullException>(
                () => new GroupInspectorPresenter(ModelStub(), null));
        }

        [Test]
        public void Constructor_givenView_setsProperty()
        {
            var model = ModelStub();
            var view = ViewStub();
            var presenter = new GroupInspectorPresenter(model, view);
            Assert.That(presenter.View, Is.EqualTo(view));
        }

        [Test]
        public void Constructor_givenView_setsPresenterPropertyOnView()
        {
            var model = ModelStub();
            var view = ViewStub();
            SetupResult.For(view.Presenter).PropertyBehavior();

            var presenter = new GroupInspectorPresenter(model, view);
            Assert.That(view.Presenter, Is.EqualTo(presenter));
        }

        [Test]
        public void InitialiseView_validPresenter_SetsSystemsAvailable()
        {
            var systems = new List<ApplicationSystem>
                              {
                                  ApplicationSystem("Desktops"),
                                  ApplicationSystem("Servers")
                              };

            var model = ModelStub();
            var view = ViewStub();

            using (mRepository.Record())
            {
                Expect.Call(model.LoadApplicationSystems())
                    .Repeat.Once()
                    .Return(systems);

                Expect.Call(() => view.SetAvailableSystems(systems))
                    .IgnoreArguments()
                    .Repeat.Once();
            }

            using (mRepository.Playback())
            {
                var presenter = new GroupInspectorPresenter(model, view);
                presenter.InitializeView();
            }
        }

        [Test]
        public void InitialiseView_validPresenter_ClearsErrors()
        {
            var systems = new List<ApplicationSystem>
                              {
                                  ApplicationSystem("Desktops"),
                                  ApplicationSystem("Servers")
                              };

            var model = ModelStub();
            var view = ViewStub();

            using (mRepository.Record())
            {
                Expect.Call(model.LoadApplicationSystems())
                    .Repeat.Once()
                    .Return(systems);

                Expect.Call(() => view.DisplayError(string.Empty, string.Empty))
                    .Repeat.Once();
            }

            using (mRepository.Playback())
            {
                var presenter = new GroupInspectorPresenter(model, view);
                presenter.InitializeView();
            }
        }

        [Test]
        public void SelectApplicationSystem_missingSystem_throwsException()
        {
            var presenter = new GroupInspectorPresenter(ModelStub(), ViewStub());
            Assert.Throws<ArgumentNullException>(
                () => presenter.SelectApplicationSystem(null));
        }

        [Test]
        public void SelectApplicationSystem_withSystem_configuresView()
        {
            var model = ModelStub();
            var view = ViewStub();
            
            var desktopSystem
                = ApplicationSystem(
                    "Desktop",
                    AccessGroup("Administrators"),
                    AccessGroup("Power Users"),
                    AccessGroup("Regular Users"));

            using (mRepository.Record())
            {
                Expect.Call(() => view.DisplayAccessGroups(null))
                    .IgnoreArguments()
                    .Repeat.Once();

                Expect.Call(model.GetUsersForAccessGroup(null))
                    .IgnoreArguments()
                    .Repeat.Times(desktopSystem.Groups.Count())
                    .Return(new List<User>());
            }

            using (mRepository.Playback())
            {
                var presenter = new GroupInspectorPresenter(model, view);
                presenter.SelectApplicationSystem(desktopSystem);
            }
        }

        [Test]
        public void SelectApplicationSystem_withSystem_displaysProgress()
        {
            var model = ModelStub();
            var view = ViewStub();

            var desktopSystem
                = ApplicationSystem(
                    "Desktop",
                    AccessGroup("Administrators"),
                    AccessGroup("Power Users"),
                    AccessGroup("Regular Users"));

            using (mRepository.Record())
            {
                Expect.Call(() => view.DisplayProgress(null, 0, 0))
                    .IgnoreArguments()
                    .Repeat.AtLeastOnce();

                Expect.Call(model.GetUsersForAccessGroup(null))
                    .IgnoreArguments()
                    .Repeat.Any()
                    .Return(new List<User>());
            }

            using (mRepository.Playback())
            {
                var presenter = new GroupInspectorPresenter(model, view);
                presenter.SelectApplicationSystem(desktopSystem);
            }
        }

        [Test]
        public void SelectApplicationSystem_GetUsersForAccessGroupThrowsException_SetsErrorMessage()
        {
            var model = ModelStub();
            var view = ViewStub();
            GroupInspectorPresenter presenter;

            var desktopSystem
                = ApplicationSystem(
                    "Desktop",
                    AccessGroup("Administrators"),
                    AccessGroup("Power Users"),
                    AccessGroup("Regular Users"));

            var error = new Exception("Something went wrong");

            using (mRepository.Record())
            {
                Expect.Call(model.GetUsersForAccessGroup(null))
                    .IgnoreArguments()
                    .Repeat.Any()
                    .Throw(error);

                Expect.Call(() => view.DisplayError(null, null))
                    .IgnoreArguments()
                    .Repeat.Once();
            }

            using (mRepository.Playback())
            {
                presenter = new GroupInspectorPresenter(model, view);
                presenter.SelectApplicationSystem(desktopSystem);
            }
        }

        [Test]
        public void SetViewStyle_withTiles_configuresView()
        {
            var model = ModelStub();
            var view = ViewStub();

            using (mRepository.Record())
            {
                Expect.Call(() => view.SetViewStyle(ViewStyle.Tiles))
                    .IgnoreArguments()
                    .Repeat.Once();
            }

            using (mRepository.Playback())
            {
                var presenter = new GroupInspectorPresenter(model, view);
                presenter.SetViewStyle(ViewStyle.Tiles);
            } 
        }

        [Test]
        public void SetViewOrder_withOrder_setsProperty()
        {
            var model = ModelStub();
            var view = ViewStub();

            var presenter = new GroupInspectorPresenter(model, view);
            presenter.SetOrder(ViewOrder.FamilyName);

            Assert.That(presenter.ViewOrder, Is.EqualTo(ViewOrder.FamilyName));
        }

        [Test]
        public void OrderUsersBy_withFamilyName_OrdersUsers()
        {
            var mandalay = UserAmyMandalay();
            var lupid = UserBettyLupid();
            var potty = UserCarminePotty();

            var natural = new List<User> {lupid, mandalay, potty};
            var expected = new List<User> {lupid, mandalay, potty};

            var model = ModelStub();
            var view = ViewStub();

            var presenter = new GroupInspectorPresenter(model, view);
            var sorted = presenter.OrderUsersBy(natural, ViewOrder.FamilyName);

            Assert.That(sorted.ToList(), Is.EqualTo(expected));
        }

        [Test]
        public void OrderUsersBy_withName_OrdersUsers()
        {
            var amy = UserAmyMandalay();
            var betty = UserBettyLupid();
            var carmine = UserCarminePotty();

            var natural = new List<User> { betty, amy, carmine };
            var expected = new List<User> { amy, betty, carmine };

            var model = ModelStub();
            var view = ViewStub();

            var presenter = new GroupInspectorPresenter(model, view);
            var sorted = presenter.OrderUsersBy(natural, ViewOrder.Name);

            Assert.That(sorted.ToList(), Is.EqualTo(expected));
        }

        [Test]
        public void OrderUsersBy_withUserName_OrdersUsers()
        {
            var underclass = UserAmyMandalay();
            var turmult = UserBettyLupid();
            var silversmith = UserCarminePotty();

            var natural = new List<User> { turmult, underclass, silversmith};
            var expected = new List<User> { silversmith, turmult, underclass };

            var model = ModelStub();
            var view = ViewStub();

            var presenter = new GroupInspectorPresenter(model, view);
            var sorted = presenter.OrderUsersBy(natural, ViewOrder.UserName);

            Assert.That(sorted.ToList(), Is.EqualTo(expected));
        }

        private User UserAmyMandalay()
        {
            return new User()
                       {
                           Name = "Amy Mandalay",
                           Surname = "Mandalay",
                           UserName = "underclass"
                       };
        }

        private User UserBettyLupid()
        {
            return new User()
                       {
                           Name = "Betty Lupid",
                           Surname = "Lupid",
                           UserName = "turmult"
                       };
        }

        private User UserCarminePotty()
        {
            return new User()
                       {
                           Name = "Carmine Potty",
                           Surname = "Potty",
                           UserName = "silversmith"
                       };
        }

        /// <summary>
        /// Create a stub of the model
        /// </summary>
        /// <returns>Model stub.</returns>
        private IGroupInspectorModel ModelStub()
        {
            return mRepository.DynamicMock<IGroupInspectorModel>();
        }

        /// <summary>
        /// Create a stub of a View
        /// </summary>
        /// <returns>New Stub</returns>
        private IGroupInspectorView ViewStub()
        {
            return mRepository.DynamicMock<IGroupInspectorView>();
        }

        private ApplicationSystem ApplicationSystem(string name, params AccessGroup[] groups)
        {
            return new ApplicationSystem()
                       {
                           Name = name,
                           Groups = groups
                       };
        }

        public AccessGroup AccessGroup(string name)
        {
            return new AccessGroup()
                       {
                           Name = name, 
                           DisplayName = name,
                           ActiveDirectoryGroupName = name
                       };
        }

        private MockRepository mRepository;
    }
}
